#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <limits.h>
#include "ap.h"

#define has_longopt(x) ((x)->longname && (x)->longname[0] != '=')
#define valid_def(x) ((x)->shortname || (x)->longname || (x)->argspec)
#define optional_def(x) ((x)->shortname || has_longopt(x))
#define positional_def(x) (!(x)->shortname && !has_longopt(x) && (x)->argspec)


static void store_defname(struct ap_res *res, const struct ap_opt *def)
{
    int off = 0;
    /* optional: -c */
    if (def->shortname) {
        res->buf[0] = '-';
        res->buf[1] = def->shortname;
        res->buf[2] = has_longopt(def) ? '/' : 0;
        off = 3;
    }
    /* optional: [-c,]--count] */
    if (has_longopt(def)) {
        res->buf[off] = res->buf[off + 1] = '-';
        off += 2;
        for (const char *s = def->longname; *s && *s != '=' && off < sizeof(res->buf) - 1; s++, off++) {
            res->buf[off] = *s;
        }
        res->buf[off] = 0;
    /* positional: COUNT */
    } else if (!def->shortname && def->longname) {
        for (const char *s = def->longname + 1; *s && off < sizeof(res->buf) - 1; s++, off++) {
            res->buf[off] = *s;
        }
        res->buf[off] = 0;
    }
}




int ap_prepare(struct ap_opt *defs, struct ap_res *res)
{
    struct ap_opt *def;
    for (def = defs; optional_def(def); def++) {
        def->maxargs &= (1 << (sizeof(def->maxargs) * CHAR_BIT / 2)) - 1;

        int rank = def->argspec & (AP_NOARG | AP_SINGLE | AP_MULTIPLE);
        if (!rank) {
            def->argspec |= AP_NOARG;
        } else if (!!(rank & AP_NOARG) + !!(rank & AP_SINGLE) + !!(rank & AP_MULTIARG) != 1) {
            store_defname(res, def);
            res->errdef = res->buf;
            res->errmsg = "definition error: multiple argument ranks given";
            return -1;
        }

        if (!(def->argspec & AP_NOARG)) {
            int type = def->argspec & (AP_BOOL | AP_INT | AP_STR);
            if (!type) {
                store_defname(res, def);
                res->errdef = res->buf;
                res->errmsg = "definition error: no argument type given";
                return -1;
            } else if (!!(type & AP_BOOL) + !!(type & AP_INT) + !!(type & AP_STR) != 1) {
                store_defname(res, def);
                res->errdef = res->buf;
                res->errmsg = "definition error: multiple argument types given";
                return -1;
            }
            int need = def->argspec & (AP_MAYBE | AP_REQUIRE);
            if (!need) {
                if (def->argspec & AP_MULTIPLE) {
                    def->argspec |= AP_MAYBE;
                } else {
                    store_defname(res, def);
                    res->errdef = res->buf;
                    res->errmsg = "definition error: no argument need given";
                    return -1;
                }
            }
        }

        if (def->longname && strlen(def->longname) > AP_MAXNAMELEN - 1) {
            res->errdef = def->longname;
            res->errmsg = "definition error: too long";
            return -1;
        }

        if (def->argspec & AP_NEED) {
            def->argspec |= AP_NEED_UNSATISFIED;
        }
    }
    for (; positional_def(def); def++) {
        /* for the rest arguments, AP_REQUIRE implies AP_NEED since there's no way to explicitly select them */
        if (def->argspec & AP_REQUIRE) {
            def->argspec |= AP_NEED;
        }
        if (def->argspec & AP_NEED) {
            def->argspec |= AP_NEED_UNSATISFIED;
        }
    }
    return 0;
}




static int store_argument(struct ap_opt *def, struct ap_res *res, const char *value)
{
    union { long l; const char *s; } val;

    if (def->argspec & AP_NOARG) {
        if (def->argspec & AP_BOOL) {
            *(unsigned *)def->val = 1;
        } else if (def->argspec & AP_INT) {
            (*(unsigned *)def->val)++;
        } else {
            store_defname(res, def);
            res->errdef = res->buf;
            res->errmsg = "argument definition can only be bool or int for argumentless options";
            return -1;
        }
    } else {
        unsigned n = def->maxargs >> (sizeof(def->maxargs) * CHAR_BIT / 2);
        unsigned max = def->maxargs & ((1 << (sizeof(def->maxargs) * CHAR_BIT / 2)) - 1);
        if (def->argspec & AP_MULTIPLE && n + 1 >= max) {
            store_defname(res, def);
            res->errdef = res->buf;
            res->errmsg = "too many arguments";
            return -1;
        }

        if (def->argspec & AP_BOOL) {
            if (!strcasecmp(value, "true") || !strcasecmp(value, "yes") || !strcmp(value, "1")) {
                val.l = 1;
            } else if (!strcasecmp(value, "false") || !strcasecmp(value, "no") || !strcmp(value, "0")) {
                val.l = 0;
            } else {
                res->errdef = value;
                res->errmsg = "not a boolean";
                return -1;
            }
            if (def->argspec & AP_MULTIPLE) {
                *((int **)def->val)[n] = val.l;
            } else {
                *(int *)def->val = val.l;
            }
        } else if (def->argspec & AP_INT) {
            char *end;
            val.l = strtol(value, &end, 10);
            if (errno == EINVAL || errno == ERANGE || *end) {
                res->errdef = value;
                res->errmsg = "not an integer";
                return -1;
            }
            if (def->argspec & AP_MULTIPLE) {
                *((long **)def->val)[n] = val.l;
            } else {
                *(long *)def->val = val.l;
            }
        } else if (def->argspec & AP_STR) {
            if (def->argspec & AP_MULTIPLE) {
                ((const char **)def->val)[n] = value;
            } else {
                *(const char **)def->val = value;
            }
        }

        if (def->argspec & AP_MULTIPLE) {
            /* store new amount of used args */
            def->maxargs &= (1 >> (sizeof(def->maxargs) * CHAR_BIT / 2)) - 1;
            def->maxargs |= (n + 1) << (sizeof(def->maxargs) * CHAR_BIT / 2);
        }
    }
    return 0;
}

static int optcmp(const char *arg, const char *opt, int *len)
{
    *len = 0;
    while (*arg && *arg != '=' && *arg == *opt) {
        arg++; opt++; (*len)++;
    }

    if (*arg && *arg != '=') {
        return 1;
    }
    if (*opt) {
        if (*opt == '=') {
            (*len)++;
            return 0;
        } else {
            return 1;
        }
    }
    *len = -1;
    return 0;
}

int ap_parse(struct ap_opt *defs, struct ap_res *res, char **args, unsigned nargs)
{
    if (ap_prepare(defs, res) < 0) {
        return -2;
    }

    /* find first positional definition */
    struct ap_opt *posdef;
    for (posdef = defs; optional_def(posdef); posdef++);

    int posonly = 0;
    for (char *arg = *args; nargs--; arg = *++args) {
        if (posonly || arg[0] != '-' || !arg[1]) {
            /* positional argument */
            if (!posdef->argspec) {
                res->errdef = arg;
                res->errmsg = "unexpected argument";
                return -1;
            }
            if (store_argument(posdef, res, arg) < 0) {
                return -1;
            }
            posdef->argspec &= ~AP_NEED_UNSATISFIED;
            if (posdef->argspec & AP_SINGLE) {
                posdef++;
            }
            continue;
        }

        arg++;
        /* -- means stop parsing optional arguments */
        if (arg[0] == '-' && !arg[1]) {
            posonly = 1;
            continue;
        }

        /* optional argument */
        int is_long = (*arg == '-');
        arg += is_long;
        do {
            int matched = 0;

            for (struct ap_opt *def = defs; optional_def(def); def++) {
                int offset = 0;
                char *value = NULL;
                int shortmatch = !is_long && def->shortname && *arg == def->shortname;
                int longmatch = is_long && has_longopt(def) && !optcmp(arg, def->longname, &offset);
                if (!shortmatch && !longmatch) {
                    continue;
                }

                matched = 1;
                if (def->argspec & AP_NOARG) {
                    if (store_argument(def, res, NULL) < 0) {
                        return -1;
                    }
                } else {
                    if (longmatch && offset && arg[offset]) {
                        value = &arg[offset];
                    }
                    if (!value && nargs) {
                        value = *++args;
                        nargs--;
                    }
                    if (!value && def->argspec & AP_REQUIRE) {
                        store_defname(res, def);
                        res->errdef = res->buf;
                        res->errmsg = "needs argument";
                        return -1;
                    }
                    if (value && store_argument(def, res, value) < 0) {
                        return -1;
                    }
                }
                def->argspec &= ~AP_NEED_UNSATISFIED;
                break;
            }

            if (!matched) {
                res->buf[0] = res->buf[1] = '-';
                if (is_long) {
                    strncpy(&res->buf[2], arg, sizeof(res->buf) - 3);
                    res->buf[sizeof(res->buf) - 1] = 0;
                } else {
                    res->buf[1] = *arg;
                    res->buf[2] = 0;
                }
                res->errdef = res->buf;
                res->errmsg = "unknown option";
                return -1;
            }
        } while (!is_long && *++arg);
    }

    struct ap_opt *def ;
    for (def = defs; valid_def(def); def++) {
        if (def->argspec & AP_NEED_UNSATISFIED) {
            store_defname(res, def);
            res->errdef = res->buf;
            res->errmsg = "required";
            return -1;
        }
    }

    return 0;
}




static void get_metavar(const struct ap_opt *def, char *buf)
{
    if (def->longname) {
        int i = 0;
        for (; i < sizeof(buf) - 1 && def->longname[i] && def->longname[i] != '='; i++) {
            buf[i] = toupper(def->longname[i]);
        }
        if (def->longname[i] != '=') {
            buf[i] = 0;
        } else {
            i++;
            strcpy(buf, def->longname + i);
        }
    } else {
        strcpy(buf, "VALUE");
    }

}

static void dump_argspec(const struct ap_opt *def)
{
    char metavar[AP_MAXNAMELEN];
    if (!(def->argspec & AP_NOARG)) {
        fputc(' ', stderr);
        /* -v [LEVEL] */
        if (def->argspec & AP_MAYBE) {
            fputc('[', stderr);
        }

        /* determine metavar from longopt, or use VALUE */
        char buf[AP_MAXNAMELEN];
        get_metavar(def, metavar);
        fputs(metavar, stderr);

        /* multiple arguments */
        if (def->argspec & AP_MULTIPLE) {
            if (def->argspec & AP_MAYBE) {
                fputs(" ...", stderr);
            } else {
                fprintf(stderr, " [%s...]", metavar);
            }
        }
        if (def->argspec & AP_MAYBE) {
            fputc(']', stderr);
        }
    }
}

void ap_usage(const struct ap_opt *defs, const char *name)
{
    /* appname [-opts...] */
    fprintf(stderr, "usage: %s", name);

    const struct ap_opt *def;
    for (def = defs; optional_def(def); def++) {
        fputc(' ', stderr);
        /* [-o OUTPUT] */
        if (!(def->argspec & AP_NEED)) {
            fputc('[', stderr);
        }
        /* prefer -o OUTPUT to --output OUTPUT */
        if (def->shortname) {
            fprintf(stderr, "-%c", def->shortname);
        } else {
            fputs("--", stderr);
            for (const char *s = def->longname; *s && *s != '='; s++) {
                fputc(*s, stderr);
            }
        }

        dump_argspec(def);
        if (!(def->argspec & AP_NEED)) {
            fputc(']', stderr);
        }
    }
    for (; positional_def(def); def++) {
        dump_argspec(def);
    }
    fputc('\n', stderr);
}

void ap_help(const struct ap_opt *defs, const char *name, const char *prologue, const char *epilogue)
{
    int maxlen = 0, len;
    char metavar[AP_MAXNAMELEN];

    /* calculate max length for padding purposes */
    for (const struct ap_opt *def = defs; valid_def(def); def++) {
        /* two leading spaces, one trailing space */
        len = 3;
        if (def->shortname) {
            len += 2;
            if (has_longopt(def)) {
                len += 2;
            }
        }
        if (has_longopt(def)) {
            for (const char *s = def->longname; *s && *s != '='; s++) {
                len++;
            }
        }
        if (!(def->argspec & AP_NOARG)) {
            len++;
            get_metavar(def, metavar);
            len += strlen(metavar);
            if ((def->argspec & AP_MAYBEARG) == AP_MAYBEARG) {
                len += 2;
            }
        }
        if (len > maxlen) {
            maxlen = len;
        }
    }
    maxlen = (maxlen + 7) & ~7;
    maxlen = maxlen < 24 ? 24 : maxlen;

    /* first, usage line; print [OPTIONS] if we have optional defs */
    const struct ap_opt *def = defs;
    fprintf(stderr, "usage: %s", name);
    if (optional_def(def)) {
        fputs(" [OPTION...]", stderr);
        for (; optional_def(def); def++);
    }
    for (; valid_def(def); def++) {
        dump_argspec(def);
    }
    fputc('\n', stderr);
    if (prologue) {
        fputs(prologue, stderr);
    }
    fputc('\n', stderr);

    /* now, extended help */
    for (const struct ap_opt *def = defs; valid_def(def); def++) {
        int llen = maxlen;

        fputs("  ", stderr);
        llen -= 2;
        if (def->shortname) {
            fprintf(stderr, "-%c", def->shortname);
            llen -= 2;
            if (has_longopt(def)) {
                fputs(", ", stderr);
                llen -= 2;
            }
        }
        if (has_longopt(def)) {
            fputs("--", stderr);
            llen -= 2;
            for (const char *s = def->longname; *s && *s != '='; s++) {
                fputc(*s, stderr);
                llen--;
            }
        }
        if (!(def->argspec & AP_NOARG)) {
            if (optional_def(def)) {
                fputc(has_longopt(def) && !((def->argspec & AP_MAYBEARG) == AP_MAYBEARG) ? '=' : ' ', stderr);
                llen--;
            }
            get_metavar(def, metavar);
            if ((def->argspec & AP_MAYBEARG) == AP_MAYBEARG) {
                fputc('[', stderr);
                llen--;
            }
            fputs(metavar, stderr);
            llen -= strlen(metavar);
            if ((def->argspec & AP_MAYBEARG) == AP_MAYBEARG) {
                fputc(']', stderr);
                llen--;
            }
        }
        fprintf(stderr, "%-*s", llen, "");
        fputs(def->desc, stderr);
        fputc('\n', stderr);
    }

    fputc('\n', stderr);
    if (epilogue) {
        fputs(epilogue, stderr);
    }
}

